/*
 * HP 54616B Oscilloscope (GPIB address: 7) β̥
 *
 * (C) 1997-01-18 K.Hirabayashi
 */

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

#define E5810A	"202.23.252.166"

unsigned char prtimg[19770];	/* HP standard raster scan method
					(512x280 dots, 19768 bytes) */
int width;			/* 512 for standard resolution */
int height;			/* 280 for standard resolution */
unsigned char img[17920];	/* 512x280 raw image data */

int adr = 7;	/* HP 54616 GPIB Address */

int bflg;	/* no view */
int Dflg;	/* DMA */
int Pflg;	/* HP standard raster method image data */
int pflg;	/* PBM format */
int rflg;	/* 640x280 raw image data (centered) */
int vflg;	/* verbose */

char *prognam;
char buf[BUFSIZ];

main(argc, argv) char **argv;
{
  int i;
  char *s;
  FILE *fp;

  prognam = argv[0];
  while (--argc > 0 && (*++argv)[0] == '-')
	for (s = argv[0]+1; *s; s++)
		switch (*s) {
		case 'a':
			if (s[1])
				s++;
			else {
				argc--; s = *++argv;

			}
			adr = atoi(s);
			s += strlen(s) - 1;
			break;
		case 'b': bflg++;
			break;
		case 'D': Dflg++;
			break;
		case 'P': Pflg++;
			break;
		case 'p': pflg++;
			break;
		case 'r': rflg++;
			break;
		case 'v': vflg++;
			break;
		default:
			Usage();
		}

  if (!isatty(0)) {	/* redirected input */
	if (Pflg) {
		if (fread(prtimg, sizeof(char), 19768, stdin) != 19768) {
			fprintf(stderr, "fread() error.\n");
			exit(1);
		}
		getimg();
	}
	else
		redImage();
	disp();
	exit(0);
  }

  if (argc > 0) {
	dump(argv[0]);
	disp();
  }
  else {
	if (!isatty(1))
		dump("-");	
	else
		dump(NULL);
  }

  exit(0);
}

/* screen dump */

dump(file) char *file;
{
  FILE *fp;
  int i, n;

  /* dump screen as HP standard raster format */
  if (gbopn(E5810A) == -1)
	error("can't open GP-IB", (char *)0);
  gblost(10000);
  gbwrt(adr, ":PRINT?", 0, 2);	/* 19768 bytes */
  n = gbred(adr, prtimg, sizeof(prtimg), 1);
  if (vflg)
	fprintf(stderr, "n(%d)(0x%02x)\n", n, gbwait(-1));
  gbcls();

  getimg();

  if (file != NULL) {
	if (strcmp(file, "-") == 0)
		fp = stdout;
	else
		fp = fopen(file, "w");
  }
  else
	fp = NULL;
  if (fp != NULL) {
	if (Pflg)
		fwrite(prtimg, sizeof(char), 19768, fp);
	else if (rflg)
		rawImage(fp);
	else
		pbmImage(fp);
	if (fp != stdout)
		fclose(fp);
  }
  disp();
}

getimg()
{
  int c, dots, bytes, i, n, p;
  unsigned char *s;

  s = prtimg;
  bytes = n = p = 0;
  while (c = *s++) {
	if (c == '\033') {
		if ((c = *s++) != '*')
			error("unexpected sequence", (char *)0);
		if ((c = *s++) == 'r') {
			if ((c = *s++) == 'A')
				;	/* start graphics */
			else if (c == 'B')
				;	/* end graphics */
			else if ('0' <= c && c <= '9') {
				for (n = 0; '0' <= c && c <= '9'; ) {
					n = n * 10 + c - '0';
					c = *s++;
				}
				if (c == 'S')
					dots = n;
				else
					error("illegal ESC * sequence (%d)", c);
			}
		}
		else if (c == 'b') {
			if ((c = *s++) < '0' || '9' < c)
				error("illegal ESC * b sequence (%d)", c);
			for (n = 0; '0' <= c && c <= '9'; ) {
				n = n * 10 + c - '0';
				c = *s++;
			}
			if (c == 'W') {
				if (bytes == 0)
					bytes = n;
				else
					if (bytes != n)
						fprintf(stderr,
"Warning: irregular line (%d)(%d)\n", bytes, n);
				for (i = 0; i < n; i++)
					img[p++] = *s++;
			}
			else
				error("illegal ESC * sequence (%d)", c);

		}
	}
	else if (c == '\n')
		break;	/* EOF */
	else
		error("non graphics data (%d)", c);
  }
  width = bytes * 8;
  height = p / bytes;
  if (vflg)
	printf("width(%d), height(%d)\n", width, height);
}

pbmImage(fp) FILE *fp;
{
  char tmps[80];
  int n;

  /* PBM RAWBITS format */
  fwrite("P4\n", sizeof(char), 3, fp);
  sprintf(tmps, "%d %d\n", width, height);
  fwrite(tmps, sizeof(char), strlen(tmps), fp);
  n = (width >> 3) * height;
  if (fwrite(img, sizeof(char), n, fp) != n)
	fprintf(stderr, "file write errormin pbmImage()\n");
  return 0;
}

redImage()
{
  int c, d, i, n;
  unsigned char *s;

  if ((c = fgetc(stdin)) == 'P') {
	if ((d = fgetc(stdin)) == '4') {
		fgets(buf, BUFSIZ, stdin);
		fgets(buf, BUFSIZ, stdin);
		sscanf(buf, "%d %d", &width, &height);
		n = (width >> 3) * height;
		if (fread(img, sizeof(char), n, stdin) != n)
			fprintf(stderr, "read error on readImage()\n");
	}
	else {
		ungetc(d, stdin);
		goto g1;
	}
  }
  else {
g1:
	ungetc(c, stdin);
	width = 512;
	height = 280;
	s = img;
	for (i = 0; i < 280; i++) {
		if ((n = fread(buf, sizeof(char), 8, stdin)) != 8)
			fprintf(stderr, "read error on readImage() 1, %d, %d\n", i, n);
		if ((n = fread(s, sizeof(char), 64, stdin)) != 64)
			fprintf(stderr, "read error on readImage() 2 %d\n", n);
		s += 64;
		if ((n = fread(buf, sizeof(char), 8, stdin)) != 8)
			fprintf(stderr, "read error on readImage() 3 %d\n", n);
	}
  }
}

/* 640 x 280 raw image */

rawImage(fp) FILE *fp;
{
  unsigned char tmps[8], *s;
  int i, n;

  s = img;
  n = width / 8;
  for (i = 0; i < 8; i++)
	tmps[i] = '\0';
  for (i = 0; i < height; i++) {
	if (fwrite(tmps, sizeof(char), 8, fp) != 8)
		fprintf(stderr, "file write errormin rawImage()\n");
	if (fwrite(s, sizeof(char), n, fp) != n)
		fprintf(stderr, "file write errormin rawImage()\n");
	s += n;
	if (fwrite(tmps, sizeof(char), 8, fp) != 8)
		fprintf(stderr, "file write errormin rawImage()\n");
  }
}

error(s, t) char *s, *t;
{
  fprintf(stderr, "%s: ", prognam);
  fprintf(stderr, s, t);
  fprintf(stderr, "\n");
  exit(1);
}

usage()
{
  fprintf(stderr, "Usage: %s [-P -r -v] [file]\n", prognam);
  exit(1);
}

FILE *
efopen(file, mod) char *file, *mod;
{
  FILE *fp;

  if ((fp = fopen(file, mod)) == NULL) {
	fprintf(stderr, "can't open %s\n", file);
	exit(1);
  }
  return fp;
}

Usage()
{
  static char *t[] = {
"ޥ	hp54600 - HP 54600 ꡼β̥",
"ʸ		hp54600 [ե饰] [ե]",
"ե饰		-a adr	GP-IB ɥ쥹 (ά 7)",
"		-P	HP standard raster method ǡ",
"		-r	640x280 饹᡼",
"		-p	PBM Υǡ (ǥե)",
"		hp54600		# ̤ɽƳǧ",
"		hp54600	file	# PBM  file ˽",
"		hp54600 -r file	# 640x280 饹᡼ file ˽",
"		hp54600 <file	# file Ƥɽ",
"		hp54600 -P <file	# file  HP standard raster",
"	hp54600 ꡼β̤פޤ줿ǡϡ", "512x280 ",
"åȤΣͥ᡼ǡˤʤޤ-r ꤷϡ 64 ɥå",
"ζɲäơFAX Ѥ 640x280 ɥåȤΥ饤᡼˳礷ޤ",
"ɸǤ PBM ΥեޤΤǡpnmtogif  GIFѴ뤳",
"ȤǤޤ-P ꤹȡ54600 ꡼פΡPRINT?ץ",
"ɤ줿HP standard raster methodץǡϤޤ",
  NULL
  };
  int i;

  for (i = 0; t[i] != NULL; i++)
	fprintf(stderr, "%s\n", t[i]);
  exit(1);
}

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#define BORDER	2

Display *dpy;
Window w, wb, wq;
GC gc;
XFontStruct *ascii, *kana, *kanji;
char *quit = "quit";

disp()
{
  Window root;
  XEvent event;
  Font font;
  char *s, *t, str[128];
  XSizeHints hints;
  XTextProperty window_name;
  XClassHint class_hint;
  XSetWindowAttributes at;

  if (bflg)
	return 0;

  if (fork() != 0)
	exit(0);

  dpy = XOpenDisplay("");

  root = XDefaultRootWindow(dpy);

  wb = XCreateSimpleWindow(dpy, root,
	0, 0, width+0, height+20,
	BORDER, BlackPixel(dpy,0), WhitePixel(dpy,0));
  wq = XCreateSimpleWindow(dpy, wb,
	0, 0, 50, 20,
	BORDER, BlackPixel(dpy,0), WhitePixel(dpy,0));
  w = XCreateSimpleWindow(dpy, wb,
	0, 20, width, height,
	BORDER, BlackPixel(dpy,0), WhitePixel(dpy,0));
  at.backing_store = Always;
  XChangeWindowAttributes(dpy, w, CWBackingStore, &at);
  XStoreName(dpy, wb, "HP 54616B");

  gc = XCreateGC(dpy, w, 0, NULL);
  XSetBackground(dpy, gc, WhitePixel(dpy,0));
  XSetForeground(dpy, gc, BlackPixel(dpy,0));
  XSelectInput(dpy, wb, ExposureMask);
  XSelectInput(dpy, wq, ExposureMask|ButtonPressMask);
  XSelectInput(dpy, w, ExposureMask);
  XMapWindow(dpy, wb);
  XMapSubwindows(dpy,wb);

  for (;;) {
	XNextEvent(dpy, &event);
	if (event.type == Expose && event.xexpose.count == 0)
		break;
  }
  XDrawString(dpy, wq, gc, 5, 15, quit, strlen(quit));

  putimg();

  for (;;) {
	XNextEvent(dpy, &event);
	switch (event.type) {
	case ButtonPress:
		if (event.xbutton.window == wq) {
			XFreeGC(dpy, gc);
			XDestroyWindow(dpy, w);
			XCloseDisplay(dpy);
			exit(0);
		}
		break;
	case Expose:
		XDrawString(dpy, wq, gc, 5, 15, quit, strlen(quit));
		break;
	}
  }
}

putimg()
{
  static int bit[] = { 128, 64, 32, 16, 8, 4, 2, 1};
  int b, j, k, n;

  b = width / 8;
  n = b * height;
  for (j = k = 0; k < n; k++)
	for (j = 0; j < 8; j++)
		if (img[k] & bit[j])
			XDrawPoint(dpy, w, gc, (k % b) * 8 + j, k / b);
}
